#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_thread.h>
#include <jack/jack.h>
#include <getopt.h>

#include "main.h"
#include "envelopes.h"
#include "find_image.h"

#include "vu_meters.h"
#include "ppm_meters.h"
#include "dpm_meters.h"

#define MET_VU  1 // The meter display type
#define MET_PPM 2
#define MET_DPM 3

void make_channel(jack_client_t *client, int i, char *port_name);
void cleanup(void);

SDL_Thread *gt;
SDL_Surface *screen;
SDL_Surface *background_image, *meter, *meter_buf;
SDL_Rect win, dest[MAX_METERS];

int num_meters;
float env[MAX_METERS];

float bias = 1.0f; // To allow for ref level changes

jack_port_t *input_ports[MAX_METERS];
jack_port_t *output_ports[MAX_METERS];
jack_client_t *client;

int main(int argc, char *argv[])
{
	SDL_Event event;
	unsigned int i;
	int opt;
	float ref_lev;
	int port_base = 1;
	unsigned int meter_type = MET_VU;
	char *meter_name = "vu";
	int columns = MAX_METERS;
	int x = 0, y = 0;
	char client_name[256];

	num_meters = argc;
	while ((opt = getopt(argc, argv, "t:r:c:h")) != -1) {
		switch (opt) {
			case 'r':
				ref_lev = atof(optarg);
				printf("Reference level: %.1fdB\n", ref_lev);
				bias = powf(10.0f, ref_lev * -0.05f);
				port_base += 2;
				break;
			case 't':
				if (!strcmp(optarg, "vu")) {
					meter_type = MET_VU;
				} else if (!strcmp(optarg, "ppm")) {
					meter_type = MET_PPM;
				} else if (!strcmp(optarg, "dpm")) {
					meter_type = MET_DPM;
				} else {
					fprintf(stderr, "Unknown meter type: %s\n", optarg);
					exit(1);
				}
				meter_name = optarg;
				port_base += 2;
				break;
			case 'c':
				columns = atoi(optarg);
				port_base += 2;
				break;
			case 'h':
				/* Force help to be shown */
				num_meters = 0;
				break;
			default:
				num_meters = 0;
				break;
		}
	}
	num_meters -= port_base;
	if (num_meters > MAX_METERS) {
		num_meters = MAX_METERS;
	}

	if (num_meters < 1) {
		fprintf(stderr, "Usage %s: [-r ref-level] [-c columns] [-t type] <port>+\n\n", argv[0]);
		fprintf(stderr, "where ref-level is the reference signal level for 0dB on the meter\n");
		fprintf(stderr, "  and type is the meter type, one of:\n");
		fprintf(stderr, "     'vu'  - classic moving needle VU meter\n");
		fprintf(stderr, "     'ppm' - PPM meter\n");
		fprintf(stderr, "     'dpm' - Digital peak meter\n");
		exit(1);
	}
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}
	//atexit(SDL_Quit);
	//atexit(cleanup);

	for (i=0; i<num_meters; i++) {
		env[i] = 0.0f;
	}

	switch (meter_type) {
		case MET_VU:
			load_graphics_vu();
			break;
		case MET_PPM:
			load_graphics_ppm();
			break;
		case MET_DPM:
			load_graphics_dpm();
			break;
	}
		
	if (columns > num_meters) {
		columns = num_meters;
	}

	/* Calculate window size */
	win.x = 0;
	win.y = 0;
	win.w = meter->w * columns;
	win.h = meter->h * (((num_meters - 1) / columns) + 1);

	screen = SDL_SetVideoMode(win.w, win.h, 24, SDL_SWSURFACE);
	if ( screen == NULL ) {
		fprintf(stderr, "Unable to set %dx%d video: %s\n", win.w, win.h, SDL_GetError());
		exit(1);
	}

	background_image = find_image("brushed-steel.png");

	dest[0].w = background_image->w;
	dest[0].h = background_image->h;
	for (x=0; x < win.w; x += background_image->w) {
		for (y=0; x < win.h; x += background_image->h) {
			dest[0].x = x;
			dest[0].y = y;
			SDL_BlitSurface(background_image, NULL, screen, dest);
		}
	}

	for (i=0, x=0, y=0; i<num_meters; i++, x++) {
		dest[i].x = meter->w * x;
		dest[i].y = meter->h * y;
		dest[i].w = meter->w;
		dest[i].h = meter->h;
		SDL_BlitSurface(meter, NULL, screen, dest+i);
		if (i % columns == columns - 1) {
			x = -1;
			y++;
		}
	}

	SDL_UpdateRects(screen, 1, &win);

	if ((client = jack_client_new ("bridge")) == 0) {
		snprintf(client_name, 255, "bridge-%d", getpid());
		if ((client = jack_client_new(client_name)) == 0) {
			fprintf(stderr, "jack server not running?\n");
			exit(1);
		}
		printf("Registering as %s\n", client_name);
		snprintf(client_name, 255, "%s meterbridge %d", meter_name, getpid());
	} else {
		printf("Registering as bridge\n");
		snprintf(client_name, 255, "%s meterbridge", meter_name);
	}
	SDL_WM_SetCaption(client_name, client_name);

	if (meter_type == MET_VU) {
		init_buffers_rms();
		jack_set_process_callback(client, process_rms, 0);
	} else {
		jack_set_process_callback(client, process_peak, 0);
	}

	switch (meter_type) {
		case MET_VU:
			gt = SDL_CreateThread(gfx_thread_vu, NULL);
			break;
		case MET_PPM:
			gt = SDL_CreateThread(gfx_thread_ppm, NULL);
			break;
		case MET_DPM:
			gt = SDL_CreateThread(gfx_thread_dpm, NULL);
			break;
	}

	if (jack_activate(client)) {
		fprintf (stderr, "cannot activate client");
		exit(1);
	}

	for (i = 0; i < num_meters; i++) {
		make_channel(client, i, argv[port_base+i]);
	}

	while (SDL_WaitEvent(&event)) {
		switch (event.type) {
			case SDL_QUIT:
				cleanup();
		}
	}

	cleanup();

	/* We can't ever get here, but it keeps gcc quiet */
	return 0;
}

/* make_channel: Do all the jack hackery needed to wire up a single channel. An
 * input port for the meter and an output port for monitoring */

void make_channel(jack_client_t *client, int i, char *port_name)
{
	char in_name[256], out_name[256];
	jack_port_t *port;
	int flags;

	snprintf(in_name, 255, "meter_%d", i+1);
	input_ports[i] = jack_port_register(client, in_name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	snprintf(out_name, 255, "monitor_%d", i+1);
	output_ports[i] = jack_port_register(client, out_name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	//XXX doesn't work properly:
	//jack_port_tie(input_ports[i], output_ports[i]);

	//fprintf(stderr, "connecting '%s' to '%s'...\n", in_name, port_name);
	port = jack_port_by_name(client, port_name);
	if (port == NULL) {
		fprintf(stderr, "Can't find port '%s'\n", port_name);
		cleanup();
	}
	flags = jack_port_flags(port);
	
	if (flags & JackPortIsInput) {
		const char **possible_ports;
		unsigned int j;

		possible_ports = jack_get_ports(client, NULL, NULL, JackPortIsOutput);
		for (j=0; possible_ports[j]; j++) {
			if (jack_disconnect(client, possible_ports[j], port_name) == 0) {
				fprintf(stderr, "Broke '%s' to '%s'\n", possible_ports[j], port_name);
				if (jack_connect(client, possible_ports[j], jack_port_name(input_ports[i]))) {
					fprintf(stderr, "Couldn't connect '%s' to '%s'\n", possible_ports[j], jack_port_name(input_ports[i]));
					cleanup();
				}
				fprintf(stderr, "Connected '%s' to '%s'\n", possible_ports[j], in_name);
			}
		}
		jack_connect(client, jack_port_name(output_ports[i]), port_name);
		fprintf(stderr, "Connected '%s' to '%s'\n", out_name, port_name);
	} else if (flags & JackPortIsOutput && jack_connect(client, port_name, jack_port_name(input_ports[i]))) {
		fprintf(stderr, "Cannot connect port '%s' to '%s'\n", port_name, in_name);
		cleanup();
	}
}

void cleanup()
{
	unsigned int i, j, k, l;
	const char **all_iports, **all_oports;
	char *disconnects[256]; //FIXME This will bite

	printf("cleanup()\n");

	/* Find all the inputs and output ports so we can check them for
	 * connections */

	all_iports = jack_get_ports(client, NULL, NULL, JackPortIsInput);
	all_oports = jack_get_ports(client, NULL, NULL, JackPortIsOutput);

	for (i=0; i<num_meters; i++) {
		k = 0;
		for (j=0; all_oports && all_oports[j]; j++) {
			/* Build a list of all the ports connected to the
			 * meters and break the connection */
			if (jack_disconnect(client, all_oports[j], jack_port_name(input_ports[i])) == 0) {
				printf("broke %s -> %s\n", all_oports[j], jack_port_name(input_ports[i]));
				disconnects[k++] = (char *)all_oports[j];
			}
		}
		for (j=0; all_iports && all_iports[j]; j++) {
			/* For every port connected to the monitor out, connect
			 * it to the ports that were connected to the
			 * equivalent meter port */
			if (jack_disconnect(client, jack_port_name(output_ports[i]), all_iports[j]) == 0) {
				printf("broke %s -> %s\n", jack_port_name(output_ports[i]), all_iports[j]);
				for (l=0; l<k; l++) {
					jack_connect(client, disconnects[l], all_iports[j]);
					printf("made %s -> %s\n", disconnects[l], all_iports[j]);
				}
			}
		}
	}

	/* Leave the jack graph */
	jack_client_close(client);

	/* Get rid of the GUI stuff */
	SDL_KillThread(gt);
	SDL_Quit();

	/* And were done */
	exit(0);
}
